/* ****************************************************************************
* Program: bsplit (binary split)
* Purpose: This program can tear a piece from a file and write this piece
*          into another file or to stdout
* Author: Eugene P. Morgunov
* Date: February 10, 2014
* Version: 0.2
* ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


void usage( char * );
/* these two functions are written in different manner deliberately */
int tear_lines();
int tear_bytes( char *, char * );

/* global variables accessible to all functions in this file */
unsigned long start;            /* starting byte or line in a source file */
unsigned long end;              /* ending byte or line in a source file */
  
FILE *source, *target;          /* source and target files */

/* -------------------------------------------------------------------------- */
/* MAIN                                                                       */
/* -------------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{
  int stts;    /* exit status of program */
  
  /* too few parameters passed to the program */
  if ( argc < 5 )
  {
    usage( argv[ 0 ] );   /* argv[ 0 ] -- program name */
  }
  
  /* atol() -- convert string to long int */
  start = atol( argv[ 2 ] );
  end = atol( argv[ 3 ] );

  /* check values of parameters */  
  if ( ! ( start > 0 && start <= end ) )
  {
    fprintf( stderr, "Parameters should conform to the condition: " \
                     "start > 0 && start <= end\n" );
    exit( 1 );
  }   
  
  source = fopen( argv[ 1 ], "r" );
  if ( source == NULL )
  {
    fprintf( stderr, "Can't open file %s\n", argv[ 1 ] );
    exit( 1 );
  } 
     
  /* if target file name is "-" then we'll open stdout */
  if ( strcmp( argv[ 4 ], "-" ) != 0 )
    target = fopen( argv[ 4 ], "w" );
  else
    /* we use fdopen() because stdout is already open at the program start */
    target = fdopen( 1 /* stdout */, "w" );
      
  if ( target == NULL )
  {
    fprintf( stderr, "Can't open file %s\n", argv[ 4 ] );
    exit( 1 );
  }  

  /* if parameter "l" was passed to the program then we count lines 
     not bytes */
  if ( argc == 6 && strcmp( argv[ 5 ], "l" ) == 0 )  
  {
    stts = tear_lines();
  }
  /* start and end of piece are byte counts */
  else
  {    
    stts = tear_bytes( argv[ 1 ], argv[ 4 ] );
  }  
  
  fclose( source );
  fclose( target );

  /* we return exit status of the program to the operating system */
  exit( stts );
}

/* -------------------------------------------------------------------------- */
/* usage                                                                      */
/* -------------------------------------------------------------------------- */
void usage( char *prg_name )
{
  int i;
  
  /* we should check if there is path name in program name and if so --
     to skip it */
       
  /* index of last symbol in program name */
  i = strlen( prg_name ) - 1;
  
  /* look for the first symbol '/' from the right */
  while ( i >= 0 && prg_name[ i ] != '/' )
    i--;
    
  /* because current symbol now is '/' or we are even before the first symbol 
     (i.e i == -1) */
  i++;
  
  printf( "Program for tearing a piece from a file\n" );
  printf( "Eugene P. Morgunov (c) 1998, 1999\n\n" );

  printf( "Usage: %s source_file start_byte end_byte target_file\n",
           &prg_name[ i ] );
  printf( "Usage: %s source_file start_line end_line target_file l\n",
           &prg_name[ i ] );
           
  printf( "If target_file is \"-\" then output goes to stdout and can be " );
  printf( "redirected\n" );
  printf( "If parameter \"l\" was passed to the program then we " \
          "count lines\n\n" );

  printf( "Example: To split file \"bigfile\" (2300123 bytes long) into " );
  printf( "two files\n" );
  printf( "\"small_1\" and \"small_2\" you should do:\n" );
  printf( "         %s bigfile 1 1400000 small_1\n", &prg_name[ i ] );
  printf( "         %s bigfile 1400001 2300123 small_2\n", &prg_name[ i ] );    
  printf( "Example: To send the end of file \"longfile\" (1000000 bytes " \
          "long) to stdout\n" );
  printf( "         with redirecting:\n" );
  printf( "         %s longfile 950000 1000000 - | more\n", &prg_name[ i ] );

  exit( 0 );
}

/* -------------------------------------------------------------------------- */
/* tearing piece measured in lines                                            */
/* -------------------------------------------------------------------------- */
int tear_lines()
{
  unsigned long i;      /* counter of read newline symbols */
  int ch;               /* current symbol */

  /* we have to count lines before starting line */     

  /* it is possible to begin from the first line */
  i = 0;

  /* at first we have to get to the starting line */
  while ( i < start - 1 )
  {
    /* read one char from a file */
    ch = fgetc( source );      

    /* end of file or some error occured */
    if ( feof( source ) || ferror( source ) )
      return ( 1 ); 
    
    /* count lines */
    if ( ch == '\n' )
      i++;
  }  

  i = start;        /* current line = starting line */

  /* now we copy characters from source stream to target stream */
  while ( i <= end )
  {
    ch = fgetc( source );      /* read one char from a file */
    if ( feof( source ) )      /* end of file encounetd */
      return ( 0 ); 
    if ( ferror( source ) )    /* some error occured */
      return ( 1 );      
    
    /* count lines */
    if ( ch == '\n' )
      i++;

    fputc( ch, target );       /* put one char to target stream */
    if ( ferror( target ) )
      return ( 1 );       
  }      
  return ( 0 );
}

/* -------------------------------------------------------------------------- */
/* tearing piece measured in bytes                                            */
/* -------------------------------------------------------------------------- */
int tear_bytes( char *source_name, char *target_name )
{
  unsigned long i;      /* counter of read and written bytes */
  int ch;               /* current symbol */
  int st;               /* status of read/write operation */
  
  /* go straightway to the beginning of a piece to be teared */ 
  fseek( source, start - 1, SEEK_SET );  /* SEEK_SET - beginning of a file */
        
  for ( i = start; i <= end; i++ )
  {
    /* we'll read and write only one byte per cycle */
    fread( &ch, 1, 1, source );
    if ( feof( source ) )   /* EOF encounters */
    {
      st = 0;
      break;
    }  
    if ( ferror( source ) )
    {
      fprintf( stderr, "Error reading file %s\n", source_name );    
      st = 1;
      break;      
    }      
    
    fwrite( &ch, 1, 1, target );
    if ( ferror( target ) )
    {
      fprintf( stderr, "Error writing file %s\n", target_name );    
      st = 1;
      break;
    }      
  }
  return ( st );
}
