/* ****************************************************************************
* Программа: динамическое распределение памяти
* Автор:     Е. П. Моргунов
* Дата:      10.02.2014
* Версия:    0.2
* ****************************************************************************/

#include <stdio.h>

// этот заголовочный файл нужен для функций malloc() и free()
#include <stdlib.h>

// этот заголовочный файл нужен для функции strlen()
#include <string.h>


int main( void )
{
  int i;    // счетчик цикла

  // Не всегда можно до запуска программы точно определить, сколько оперативной
  // памяти потребуется при выполнении программы. Поэтому нужно иметь
  // возможность получать необходимое количество памяти уже в процессе
  // выполнения программы. Такой механизм в языке C существует. Основной
  // функцией для получения некоторого объема оперативной памяти является
  // функция malloc().
  
  // Выделим память под массив целых чисел и под символьную строку. Когда
  // операционная система выделяет программе блок памяти, она дает также
  // адрес первого байта этого блока памяти. Этот адрес программа должна
  // запомнить, чтобы с его помощью получить доступ к выделенному блоку памяти.
  // Запомнить адрес можно только в переменной, которая является указателем.
  // Поэтому прежде чем запрашивать у операционной системы блок памяти, нужно
  // объявить переменную-указатель.

  int *numbers;    // указатель на тип int (целые числа)
  char *string;    // указатель на тип char

  // выполним непосредственый запрос на динамическое распределение памяти
  // для 10 целых чисел; в этом операторе мы использовали следующие приемы:
  // -- параметром функции malloc() является число запрашиваемых байтов, однако
  // нам нужно разместить 10 значений типа int. Поскольку в принципе в разных
  // операционных системах значение этого типа может занимать не только 4 байта
  // в памяти, но и другое число байтов, то более надежным подходом будет
  // использование операции sizeof, т. к. компилятор "знает" наверняка, какова
  // длина типа данных int. Таким образом, требуемое число байтов мы получаем 
  // путем умножения размера одного значения требуемого типа на число таких
  // значений;
  // -- функция malloc() возвращает адрес выделенного нашей программе блока
  // памяти в виде указателя на так называемый тип void. Мы же должны 
  // преобразовать такой указатель в указатель именно на тип int. Эта 
  // операция называется ПРИВЕДЕНИЕМ типа и выполняется путем размещения
  // префикса ( int * ) перед вызовом функции malloc().
  numbers = ( int * )malloc( sizeof( int ) * 10 );

  // нужно обязательно проверять успешность работы функции malloc(),
  // т. к. в принципе она может и НЕ выделить требуемый объем памяти,
  // а возвратить в этом случае указатель NULL
  if ( numbers == NULL )
  {
    puts( "Не могу выделить память для массива целых чисел" );
    exit( 1 );     // "аварийное" завершение программы
  }

  // память для символьной строки распределяется аналогично, однако, 
  // использовать операцию sizeof нет необходимости, т. к. одно значение
  // типа char занимает один байт памяти; но с другой стороны, при 
  // распределении памяти для символьных строк нужно помнить о том, что
  // каждая такая строка должна завершаться нулевым символом '\0'. Именно
  // он служит концом строки, именно на него ориентируется функция strlen(),
  // которая возвращает длину строки. Поэтому, если мы хотим получить блок
  // памяти какого-то точного размера, предназначенный для хранения заранее
  // известного нам текста, то мы должны запросить у функции malloc() на 1 байт
  // больше -- для конечного символа '\0'. Если же мы запрашиваем блок памяти,
  // как говорится, с запасом, то такой точный расчет делать уже не нужно.
  string = ( char * )malloc( 6 + 1 );   // будем хранить строку "Russia"

  if ( string == NULL )
  {
    puts( "Не могу выделить память для строки" );
    exit( 1 );
  }

  // заполним массив случайными значениями в цикле
  // константа RAND_MAX определена в заголовочном файле stdlib.h
  // ПРИМЕЧАНИЕ. Поскольку в цикле выполняется всего один оператор, то можно
  //             не использовать скобки {} для ограничения тела цикла.
  printf( "Заполним массив случайными значениями в диапазоне от 0 до %d\n",
          RAND_MAX );

  // Для того чтобы при каждом запуске программы не выводилась одна и та же
  // последовательность псевдослучайных чисел, инициализируем генератор
  // случайных чисел. Используем приведение к типу unsigned int, т.к.
  // функция time() возвращает значение типа time_t.
  srand( ( unsigned int )time( 0 ) );

  for ( i = 0; i < 10; i++ )
    numbers[ i ] = rand();

  // слово "Россия" запишем в выделенную область памяти "вручную"
  string[ 0 ] = 'R';
  string[ 1 ] = 'u';
  string[ 2 ] = 's';
  string[ 3 ] = 's';
  string[ 4 ] = 'i';
  string[ 5 ] = 'a';
  string[ 6 ] = '\0';      // обязательный нулевой символ

  // выведем числа из массива
  for ( i = 0; i < 10; i++ )
    printf( "%d\n", numbers[ i ] );

  // выведем строку как единое целое
  printf( "\n%s\n\n", string );

  // выведем строку посимвольно
  // обратите внимание, что спецификатор формата -- %с
  for ( i = 0; i < strlen( string ); i++ )
    printf( "%c\n", string[ i ] );

  // Блоки памяти, полученные посредством динамического распределения,
  // необходимо возвратить операционной системе. Для этого служит функция
  // free(), которая вызывается с параметром, в качестве которого используется
  // адрес блока памяти, полученного с помощью функции malloc().
  free( numbers );
  free( string );

  return 0;
}
